from ase.io.lammpsrun import read_lammps_dump
from ase.build import bulk
from ase import Atoms
from ase.io import write as ase_write
from ovito.io import import_file, export_file
from scipy.spatial import cKDTree
import numpy as np
import tempfile, os
from typing import List

class WSMet:
    """
    Clase para generar una red perfecta a partir de un dump defectuoso y detectar vacancias.
    """

    def __init__(self,
                 defect_dump_path: str,
                 lattice_type: str = 'bcc',
                 element: str = 'Fe',
                 tolerance: float = 0.5,
                 perfect_dump_path: str = 'inputs/red_perfecta.dump',
                 vacancies_xyz_path: str = 'vacancias_detectadas.xyz'):
        """
        Inicializa el analizador.

        Parámetros:
        - defect_dump_path: ruta al dump LAMMPS defectuoso.
        - lattice_type: 'bcc' o 'fcc'.
        - element: símbolo del elemento, e.g. 'Fe'.
        - tolerance: distancia umbral para detectar vacancias (Å).
        - perfect_dump_path: ruta de salida para el dump perfecto.
        - vacancies_xyz_path: ruta de salida para el xyz de vacancias.
        """
        self.defect_dump_path = defect_dump_path
        self.lattice_type = lattice_type.lower()
        self.element = element
        self.tolerance = tolerance
        self.perfect_dump_path = perfect_dump_path
        self.vacancies_xyz_path = vacancies_xyz_path

        # Cargar estructura defectuosa
        self.at_defect = read_lammps_dump(self.defect_dump_path, index=-1)

    def _compute_lattice_parameter(self, atoms: Atoms) -> float:
        """Calcula el parámetro de red a partir de la densidad atómica."""
        cell = atoms.get_cell()
        vol = cell.volume
        n = len(atoms)
        natoms_cell = 2 if self.lattice_type == 'bcc' else 4
        density = n / vol
        cell_vol = natoms_cell / density
        return cell_vol ** (1/3)

    def _compute_ideal_distance(self, a: float) -> float:
        """Calcula la distancia ideal al primer vecino en la red."""
        if self.lattice_type == 'bcc':
            return np.sqrt(3) / 2 * a
        elif self.lattice_type == 'fcc':
            return a / np.sqrt(2)
        else:
            raise ValueError("Tipo de red no reconocido: usar 'bcc' o 'fcc'.")

    def generate_perfect_atoms(self) -> Atoms:
        """Genera el objeto Atoms de la red perfecta sin vacancias."""
        # calcular parámetro de red
        a = self._compute_lattice_parameter(self.at_defect)
        # construir celda unitaria perfecta
        red_unit = bulk(name=self.element, crystalstructure=self.lattice_type, a=a, cubic=True)
        # repetir para llenar la celda original
        reps = np.ceil(self.at_defect.get_cell().lengths() / a).astype(int)
        red = red_unit * reps.tolist()
        # escalar y alinear
        red.set_cell(self.at_defect.get_cell(), scale_atoms=True)
        red.set_pbc([True, True, True])
        centro_def = np.mean(self.at_defect.get_positions(), axis=0)
        centro_red = np.mean(red.get_positions(), axis=0)
        red.translate(centro_def - centro_red)
        return red

    def export_perfect_dump(self, atoms_perfect: Atoms) -> None:
        """Exporta la estructura perfecta a un archivo LAMMPS dump usando OVITO."""
        with tempfile.NamedTemporaryFile(suffix=".xyz", delete=False) as tmp:
            ase_write(tmp.name, atoms_perfect, format="xyz")
            tmp_path = tmp.name
        pipeline = import_file(tmp_path)
        export_file(pipeline, self.perfect_dump_path, "lammps/dump",
                    columns=["Particle Identifier","Particle Type","Position.X","Position.Y","Position.Z"])
        os.remove(tmp_path)

    def detect_vacancies(self, atoms_perfect: Atoms) -> List[np.ndarray]:
        """Detecta vacancias comparando la red perfecta con la defectuosa."""
        pos_perfect = atoms_perfect.get_positions()
        pos_defect = self.at_defect.get_positions()
        tree = cKDTree(pos_defect)
        vac_positions = []
        for pos in pos_perfect:
            dist, _ = tree.query(pos, k=1)
            if dist > self.tolerance:
                vac_positions.append(pos)
        # exportar XYZ si hay vacancias
        if vac_positions:
            vac_atoms = Atoms(positions=vac_positions,
                              symbols=[atoms_perfect[0].symbol]*len(vac_positions))
            vac_atoms.set_cell(atoms_perfect.get_cell())
            vac_atoms.set_pbc([True, True, True])
            ase_write(self.vacancies_xyz_path, vac_atoms)
        return vac_positions

    def run(self) -> List[np.ndarray]:
        """Ejecuta flujo completo: genera red perfecta, exporta .dump y detecta vacancias."""
        #print(f"🔧 Generando red perfecta para '{self.defect_dump_path}'")
        atoms_perfect = self.generate_perfect_atoms()
        #print(f"💾 Exportando red perfecta a '{self.perfect_dump_path}'")
        self.export_perfect_dump(atoms_perfect)
        print("🔍 Detectando vacancias...")
        vacs = self.detect_vacancies(atoms_perfect)
        print(f"✅ Vacancies encontradas: {len(vacs)}")
        return vacs

# Ejemplo de uso:
# analyzer = VacancyAnalyzer("inputs/void_15.dump", "bcc", "Fe", tolerance=0.5)
# vac_positions = analyzer.run()



import os
import json
import math
import numpy as np

class KeyFilesSeparator:
    def __init__(self, config, clusters_json_path):
        self.config = config
        self.cluster_tolerance = config['cluster tolerance']
        self.clusters_json_path = clusters_json_path
        self.lista_clusters_final = []
        self.lista_clusters_criticos = []
        self.num_clusters = self.cargar_num_clusters()

    def cargar_num_clusters(self):
        if not os.path.exists(self.clusters_json_path):
            return 0
        with open(self.clusters_json_path, "r", encoding="utf-8") as f:
            datos = json.load(f)
        num = datos.get("num_clusters", 0)
        return num

    def extraer_coordenadas(self, file_path):
        coordenadas = []
        try:
            with open(file_path, 'r', encoding='utf-8') as f:
                lines = f.readlines()
        except FileNotFoundError:
            return coordenadas
        start_index = None
        for i, line in enumerate(lines):
            if line.strip().startswith("ITEM: ATOMS"):
                start_index = i + 1
                break
        if start_index is None:
            return coordenadas
        for line in lines[start_index:]:
            parts = line.split()
            if len(parts) < 6:
                continue
            try:
                x = float(parts[2])
                y = float(parts[3])
                z = float(parts[4])
                coordenadas.append((x, y, z))
            except ValueError:
                continue
        return coordenadas

    def calcular_centro_de_masa(self, coordenadas):
        arr = np.array(coordenadas)
        if arr.size == 0:
            return None
        centro = arr.mean(axis=0)
        return tuple(centro)

    def calcular_dispersion(self, coordenadas, centro_de_masa):
        if coordenadas is None or (hasattr(coordenadas, '__len__') and len(coordenadas) == 0) or centro_de_masa is None:
            return [], 0
        distancias = []
        cx, cy, cz = centro_de_masa
        for (x, y, z) in coordenadas:
            d = math.sqrt((x - cx) ** 2 + (y - cy) ** 2 + (z - cz) ** 2)
            distancias.append(d)
        dispersion = np.std(distancias)
        return distancias, dispersion

    def construir_matriz_coordenadas(self, archivo):
        coords = self.extraer_coordenadas(archivo)
        matriz = []
        for (x, y, z) in coords:
            matriz.append([x, y, z, 0])
        return np.array(matriz)

    def separar_archivos(self):
        for i in range(1, self.num_clusters + 1):
            ruta_archivo = f"outputs/dump/key_area_{i}.dump"
            coords = self.extraer_coordenadas(ruta_archivo)
            centroide = self.calcular_centro_de_masa(coords)
            distancias, dispersion = self.calcular_dispersion(coords, centroide)
            if dispersion > self.cluster_tolerance:
                self.lista_clusters_criticos.append(ruta_archivo)
            else:
                self.lista_clusters_final.append(ruta_archivo)

    def exportar_listas(self, output_path):
        datos_exportar = {
            "clusters_criticos": self.lista_clusters_criticos,
            "clusters_final": self.lista_clusters_final
        }
        with open(output_path, "w", encoding="utf-8") as f:
            json.dump(datos_exportar, f, indent=4)

    def run(self):
        self.separar_archivos()
        self.exportar_listas("outputs/json/key_archivos.json")


import json
import copy
import os
import csv
import numpy as np
from scipy.spatial import ConvexHull
import pandas as pd
from ovito.io import import_file, export_file
from ovito.modifiers import (
    ExpressionSelectionModifier,
    DeleteSelectedModifier,
    ClusterAnalysisModifier,
    ConstructSurfaceModifier,
    InvertSelectionModifier
)
from vfscript.training.training_fingerstyle import DumpProcessor, StatisticsCalculator
from vfscript.training.utils import resolve_input_params_path

class AtomicGraphGenerator:
    def __init__(self, json_params_path: str = None):
        # --- Parámetros de input_params.json -------------
        if json_params_path is None:
            json_params_path = resolve_input_params_path("input_params.json")
        cfg = json.load(open(json_params_path))["CONFIG"][0]
        self.input_path = cfg['relax']
        self.cutoff    = cfg['cutoff']
        self.radius    = cfg['radius']
        self.smoothing = cfg['smoothing_level_training']
        self.iterations= cfg['max_graph_variations']
        self.max_nodes = cfg['max_graph_size']
        # -------------------------------------------------

        self.pipeline = import_file(self.input_path, multiple_frames=True)

        # Prepara JSON
        self.records = []

        # Rutas y header
        self.csv_path = "outputs/csv/finger_data.csv"
        os.makedirs(os.path.dirname(self.csv_path), exist_ok=True)

        header = [
            "vacancys",  # tú pusiste 'length', pon aquí el nombre real
            "N", "mean", "std",
            "skewness", "kurtosis", "Q1", "median", "Q3", "IQR"
        ] + [f"hist_bin_{i}" for i in range(1,11)]

        # Sólo escribir header si el archivo NO existe o está vacío
        if not os.path.exists(self.csv_path) or os.path.getsize(self.csv_path) == 0:
            with open(self.csv_path, 'w', newline='', encoding='utf-8') as csvfile:
                writer = csv.writer(csvfile)
                writer.writerow(header)
    def run(self):
        for variation_idx in range(self.iterations):
            for graph_size in range(1, self.max_nodes + 1):
                ids, _ = self._generate_graph(graph_size)
                expr = " || ".join(f"ParticleIdentifier=={pid}" for pid in ids)

                area, volume, count, dump_path = self._export_and_dump(
                    expr,
                    graph_size,
                    variation_idx
                )
                # 2) extraer normas y estadísticas
                proc = DumpProcessor(dump_path)
                proc.read_and_translate()
                proc.compute_norms()
                stats = StatisticsCalculator.compute_statistics(proc.norms)

                # 3) agregar al JSON interno
                rec = {
                    "surface_area": area,
                    "filled_volume": volume,
                    "vacancys":     len(ids),
                    "cluster_size": count,
                    **stats
                }
                self.records.append(rec)

                # 4) escribir línea en el CSV
                row = [
                    len(ids),
                    stats['N'],
                    stats['mean'],
                    stats['std'],
                    stats['skewness'],
                    stats['kurtosis'],
                    stats['Q1'],
                    stats['median'],
                    stats['Q3'],
                    stats['IQR']
                ] + [stats[f"hist_bin_{i}"] for i in range(1,11)]
                with open(self.csv_path, 'a', newline='', encoding='utf-8') as csvfile:
                    writer = csv.writer(csvfile)
                    writer.writerow(row)

        # Guardar JSON completo
        os.makedirs("outputs/json", exist_ok=True)
        with open("outputs/json/training_graph.json", "w", encoding='utf-8') as f:
            json.dump(self.records, f, indent=4)
        print(f"✔️ JSON guardado en outputs/json/training_graph.json")
        print(f"✔️ CSV guardado en       {self.csv_path}")

    def _generate_graph(self, length: int):
        data    = self.pipeline.compute()
        pos     = data.particles.positions.array
        ids_arr = data.particles['Particle Identifier'].array
        N       = len(pos)
        start   = np.random.choice(N)
        coords  = [pos[start]]
        ids     = [int(ids_arr[start])]
        current = coords[0]
        rem_set = set(range(N)) - {start}

        while len(coords) < length and rem_set:
            rem    = np.array(list(rem_set))
            dists  = np.linalg.norm(pos[rem] - current, axis=1)
            order  = np.argsort(dists)
            cands  = rem[order[:2]] if len(order) > 1 else rem[order]
            choice = np.random.choice(cands)
            coords.append(pos[choice])
            ids.append(int(ids_arr[choice]))
            current = pos[choice]
            rem_set.remove(choice)

        return ids, coords

    def _export_and_dump(self, expr: str,i: int,a:int):
        p = copy.deepcopy(self.pipeline)
        p.modifiers.append(ExpressionSelectionModifier(expression=expr))
        p.modifiers.append(DeleteSelectedModifier())
        p.modifiers.append(ConstructSurfaceModifier(
            radius=self.radius,
            smoothing_level=self.smoothing,
            select_surface_particles=True
        ))
        p.modifiers.append(InvertSelectionModifier())
        p.modifiers.append(DeleteSelectedModifier())
        p.modifiers.append(ClusterAnalysisModifier(cutoff=self.cutoff, unwrap_particles=True))

        data = p.compute()
        pts  = data.particles.positions.array
        count= len(pts)
        if count >= 4:
            hull   = ConvexHull(pts)
            area   = hull.area
            volume = hull.volume
        else:
            area, volume = 0.0, 0.0

        dump_dir  = "outputs/dump"
        os.makedirs(dump_dir, exist_ok=True)
        dump_path = os.path.join(dump_dir, f"graph_{i}_{a}.dump")
        export_file(
            p, dump_path, 'lammps/dump',
            columns=[
                'Particle Identifier','Particle Type',
                'Position.X','Position.Y','Position.Z'
            ]
        )
        p.modifiers.clear()
        return area, volume, count, dump_path





import warnings
warnings.filterwarnings('ignore', message='.*OVITO.*PyPI')

import ovito._extensions.pyscript
# … resto de imports de OVITO …
import pandas as pd
import os
from .surface_processor.surface_processor import SurfaceProcessor
from .surface_processor.cluster_dump_processor import ClusterDumpProcessor
from .cluster_processing.cluster_processor import ClusterProcessor, ClusterProcessorMachine
from .cluster_processing.key_files_separator import KeyFilesSeparator
from .cluster_processing.export_cluster_list import ExportClusterList
from .training.cristal_structure_gen import CrystalStructureGenerator
from .training.training_surface import HSM
import os
import json
from .cluster_processing.cluster_macth import DumpProcessorFinger, StatisticsCalculatorFinger, JSONFeatureExporterFinger
from .training.training_processor import TrainingProcessor
from .predictors.vacancy_predictors import (
    VacancyPredictorRF,
    XGBoostVacancyPredictor,
    VacancyPredictor,
    VacancyPredictorMLP
)

from pathlib import Path
from .training.utils import load_json_data, resolve_input_params_path
from .runner.finger_runner import WinnerFinger
import json
from .training.training_btree import BehaviorTreeModel
from  .predictors.vacancy_predictors_classified import VacancyModelTrainer
from .predictors.mach_finger import FingerprintVacancyAssigner
from .training.training_defect_fingerstyle import ClusterFeatureExporter
from .training.training_btr_assing import ImprovedVacancyClassifier
from .training.training_graph import AtomicGraphGenerator
from . runner.deformation_analyzer import DeformationAnalyzer
from . runner.ws_predictor import WSMet


# modifiers/utils/io_utils.py

import os
import json
import pandas as pd

def load_json_data(json_path: str) -> pd.DataFrame:
    """
    Lee un JSON y devuelve un DataFrame de pandas.
    """
    with open(json_path, "r", encoding="utf-8") as f:
        data = json.load(f)
    return pd.DataFrame(data)

def resolve_input_params_path(module_file: str, json_name: str = "input_params.json") -> str:
    """
    Busca primero 'input_params.json' en la misma carpeta que module_file.
    Si no existe, sube un nivel y busca allí.
    """
    carpeta_actual = os.path.dirname(module_file)
    ruta_local = os.path.join(carpeta_actual, json_name)
    if os.path.exists(ruta_local):
        return ruta_local

    carpeta_padre = os.path.dirname(carpeta_actual)
    ruta_padre = os.path.join(carpeta_padre, json_name)
    if os.path.exists(ruta_padre):
        return ruta_padre

    raise FileNotFoundError(f"No se encontró '{json_name}' en {carpeta_actual} ni en {carpeta_padre}")



from setuptools import setup, find_packages

setup(
    name='vacancycalculator',
    version='0.3.4.9',
    author='E.Bringa-S.Bergamin-SiMaF',
    author_email='santiagobergamin@gmail.com',
    license='MIT',
    description='Defect analysis and vacancy calculation for materials science',
    url='https://github.com/TiagoBe0/VFScript-SiMaF',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    install_requires=[
        'numpy',
        'scipy',
        'pandas',
        'scikit-learn',
        'xgboost',
        'ase',
        'ovito'
    ],
    include_package_data=True,
)
